@@ -1,11 +1,15 @@
 '''
 Module containing easy_db DataBase class.
 '''
-from typing import Union, List
+from typing import Union, List, Set
 import sqlite3
-import pyodbc
+try:
+    import pyodbc
+except Exception:
+    pass
 import os
 import time
+import math
 import random
 from functools import lru_cache
 import tqdm
@@ -18,9 +22,10 @@ from . import util
 
 class DataBase():
 
-    def __init__(self, db_location_str: str='', create_if_none: bool=True) -> None:
+    def __init__(self, db_location_str: str='', create_if_none: bool=True):
         self.db_location_str = db_location_str
         self._pull_cache: dict = {}
+        self._conn_cache = None
 
         self.db_type = self._find_db_type()
         if self.db_type == 'ACCESS':
@@ -64,9 +69,9 @@ class DataBase():
             return 'Database not recognized!'
 
 
-    def _connection_sqlite(self, also_cursor: bool=False, create_if_none: bool=False):
+    def _connection_sqlite(self, also_cursor: bool=False, create_if_none: bool=False, **kwargs):
         '''
-        Return a connection object to the Sqlite Database.
+        Return a connection object to the SQLite Database.
         '''
         db_file_exists = True if os.path.isfile(self.db_location_str) else False
         if db_file_exists or create_if_none:
@@ -80,9 +85,12 @@ class DataBase():
             print('Please first create this database or specify create_if_none=True.')
 
 
-    def _connection_access(self, also_cursor: bool=False):
+    def _connection_access(self, also_cursor: bool=False, cache_conn: bool=False, **kwargs):
         '''
         Return a connection object to the Access Database.
+        cache_conn kwarg specified as True will cache the same access connection to
+        greatly improve performance if this a collection is called for many times.
+        (For example many db.update(..., cache_conn=True) calls in a row.)
         '''
         if not os.path.isfile(self.db_location_str):
             if '.accdb' in self.db_location_str and os.path.isfile(self.db_location_str.replace('.accdb', '.mdb')):
@@ -94,16 +102,53 @@ class DataBase():
             raise FileNotFoundError(error_str)
 
         absolute_path = os.path.abspath(self.db_location_str)  # NEED AN ABSOLUTE PATH FOR PYODBC!!!
-        conn = pyodbc.connect(
-            r"Driver={Microsoft Access Driver (*.mdb, *.accdb)};" +
-            r'Dbq=' + absolute_path + ';')
+
+        # Shortcut to cached connection if desired
+        if cache_conn and self._conn_cache:
+            if also_cursor:
+                return self._conn_cache, self._conn_cache.cursor()
+            else:
+                return self._conn_cache
+
+        # If no longer calling for a cached connection, try closing out if exists as best practice
+        if not cache_conn:
+            try:
+                self._conn_cache.close()
+                self._conn_cache = None
+            except Exception:
+                pass
+
+        # try to connect a few times if first pass fails
+        # may occur if the Access locking/unlocking process is taking longer than usual
+        conn, tries = None, 0
+        while conn is None:
+            try:
+                tries += 1
+                conn = pyodbc.connect(
+                    r'Driver={Microsoft Access Driver (*.mdb, *.accdb)};' +
+                    r'Dbq=' + absolute_path + ';')
+            except pyodbc.Error:
+                time.sleep(0.7)  # time delay so Access can hopefully get unlocked
+            if tries > 5:
+                break
+
+        # now try again one more time to get the pyodbc error message/traceback
+        if conn is None:
+            conn = pyodbc.connect(
+                r'Driver={Microsoft Access Driver (*.mdb, *.accdb)};' +
+                r'Dbq=' + absolute_path + ';')
+
+        # If want to cache but didn't have saved previously for early return, save conn now for later calls
+        if cache_conn:
+            self._conn_cache = conn
+
         if also_cursor:
             return conn, conn.cursor()
         else:
             return conn
 
 
-    def _connection_sql_server(self, also_cursor: bool=False):
+    def _connection_sql_server(self, also_cursor: bool=False, **kwargs):
         '''
         Return a connection object to the SQL Server Database.
         '''
@@ -158,7 +203,23 @@ class DataBase():
             print(f'Current database is: {self.db_type}')
 
 
-    def pull(self, tablename: str, columns='all', fresh=False, progress_handler=None) -> list:
+    def execute(self, sql: str, parameters: list=[]):
+        '''
+        Shortcut for cursor.execute().
+
+        Return similar to how cursor.execute() behaves.  That is, if a SELECT query, return selected data.
+        If no return data, return None.
+        '''
+        with self as cursor:
+            query = cursor.execute(sql, parameters)
+            try:
+                result = query.fetchall()
+                return result if result else None
+            except:
+                return None
+
+
+    def pull(self, tablename: str, columns='all', fresh=False, progress_handler=None, cache_conn: bool=False) -> list:
         '''
         "SELECT *" query for full table as specified from tablename.
         ALSO WORKS for an Access Select query named tablename!
@@ -180,28 +241,38 @@ class DataBase():
 
         Return list of dicts for rows with column names as keys.
         '''
-        if tablename not in self.table_names() + self.query_names():
-            print(f'Table or query "{tablename}" not found.  Pull aborted.')
-            return []
-
         if fresh:
-            self._pull_cache = {}
-            return self.pull(tablename, columns)
+            self._clear_pull_cache(tablename)  # clear cache for this table
+            return self.pull(tablename, columns, progress_handler=progress_handler)
+
         else:
-            # check for questionable table/column names
-            for name in [tablename] + list(columns):
-                if not util.name_clean(name):
+            if columns == 'all':
+                requested_data_key = tablename
+            else:
+                requested_data_key = f'{tablename}_' + '_'.join(sorted(columns))  # key string for caching db pulls in dict
+
+            try:
+                return self._pull_cache[requested_data_key]
+
+            except KeyError:
+                # check for questionable table/column names
+                for name in [tablename] + list(columns):
+                    if not util.name_clean(name):
+                        return []
+
+                conn, cursor = self.connection(also_cursor=True, cache_conn=cache_conn)
+
+                # ensure specified tablename is a valid table (or query possibly in Access)
+                if tablename not in self.table_names(existing_cursor=cursor) + self.query_names(existing_cursor=cursor):
+                    print(f'Table or query "{tablename}" not found.  Pull aborted.')
                     return []
 
-            requested_data_key = f'{tablename}_' + '_'.join(sorted(columns))  # key string for caching db pulls in dict
-            if requested_data_key not in self._pull_cache:
                 if columns == 'all':
                     sql = f'SELECT * FROM "{tablename}";'
                 elif isinstance(columns, str):
                     columns = [columns]  # convert to list for a single user-provided column string
                 else:
                     sql = f'SELECT {", ".join(columns)} FROM "{tablename}";'
-                conn, cursor = self.connection(also_cursor=True)
 
                 if progress_handler is not None:
                     if self.db_type == 'SQLITE':  # progress_handler only currently working for sqlite
@@ -209,9 +280,17 @@ class DataBase():
                     else:
                         print('progress_handler is only available for use with a SQLite database.')
 
-                self._pull_cache[requested_data_key] = util.list_of_dicts_from_query(cursor, sql, tablename, self.db_type)
-                conn.close()
-            return self._pull_cache[requested_data_key]
+                self._pull_cache[requested_data_key] = util.list_of_dicts_from_query(cursor, sql, tablename, self.db_type, columns=columns if isinstance(columns, list) else [])
+                if not cache_conn:
+                    conn.close()
+                return self._pull_cache[requested_data_key]
+
+
+    def _clear_pull_cache(self, tablename) -> None:
+        '''Fully clear pull cache for all keys related to the specified table.'''
+        for key in list(self._pull_cache.keys()):
+            if tablename in key:
+                self._pull_cache.pop(key, None)
 
 
     def pull_where(self, tablename: str, condition: str, columns='all') -> list:
@@ -221,20 +300,24 @@ class DataBase():
         '''
         if columns == 'all':
             sql = f'SELECT * FROM {tablename} WHERE {condition};'
+        elif isinstance(columns, str):
+            columns = [columns]  # convert to list for a single user-provided column string
         elif isinstance(columns, list):  # list of columns to pull
             sql = f'SELECT {", ".join(columns)} FROM {tablename} WHERE {condition};'
         else:
             print('Columns kwarg for .pull_where must be a list of column names.')
-        conn, cursor = self.connection(also_cursor=True)
-        data = util.list_of_dicts_from_query(cursor, sql, tablename, self.db_type)
-        conn.close()
-        return data
+
+        with self as cursor:
+            return util.list_of_dicts_from_query(cursor, sql, tablename, self.db_type, columns=columns if isinstance(columns, list) else [])
 
 
-    def pull_where_id_in_list(self, tablename: str, id_col: str, match_values: list, columns='all', use_multip: bool=False, progressbar: bool=True) -> list:
+    def pull_where_id_in_list(self, tablename: str, id_col: str, match_values: list, columns='all', use_multip: bool=False, progressbar: bool=False) -> list:
         '''
         Pulls all data from table where id_col value is in the provided match_values.
         '''
+        if isinstance(columns, str) and columns != 'all':
+            columns = [columns]
+
         @lru_cache(maxsize=4)
         def sql_str(subset_len: int) -> str:
             if columns == 'all':
@@ -245,47 +328,65 @@ class DataBase():
         if progressbar:
             pbar = tqdm.tqdm(total=len(match_values))
 
-        conn, cursor = self.connection(also_cursor=True)
-        data: list = []
-        while len(match_values) > 0:
-            subset = match_values[:100]
-            sql = sql_str(len(subset))
-            data.extend(util.list_of_dicts_from_query(cursor, sql, tablename, self.db_type, subset))
-            match_values = match_values[100:]
+        with self as cursor:
+            data: list = []
+            while len(match_values) > 0:
+                subset = match_values[:100]
+                sql = sql_str(len(subset))
+                data.extend(util.list_of_dicts_from_query(cursor, sql, tablename, self.db_type, subset, columns=columns if isinstance(columns, list) else []))
+                match_values = match_values[100:]
+                if progressbar:
+                    pbar.update(100)
             if progressbar:
-                pbar.update(100)
+                pbar.update(len(match_values) % 100)
 
-        if progressbar:
-            pbar.update(len(match_values) % 100)
-
-        conn.close()
         return data
 
 
+    def distinct_values(self, tablename: str, columnname: str) -> Set:
+        '''
+        Pull and return set of unique values in given column of given table.
+        '''
+        with self as cursor:
+            cursor.execute(f'SELECT DISTINCT {columnname} FROM {tablename};')
+            return set(tup[0] for tup in cursor.fetchall())
+
+
     @lru_cache(maxsize=1)
-    def table_names(self) -> list:
+    def table_names(self, existing_cursor=None) -> list:
         '''
         Return sorted list of all tables in the database.
         '''
-        conn, cursor = self.connection(also_cursor=True)
-        if self.db_type == 'SQLITE':
-            tables = [tup[0] for tup in cursor.execute("SELECT name FROM sqlite_master WHERE type='table';").fetchall()]
-        elif self.db_type == 'ACCESS':
-            tables = [tup[2] for tup in cursor.tables() if tup[3] == 'TABLE']
+        if existing_cursor:
+            if self.db_type == 'SQLITE':
+                tables = [tup[0] for tup in existing_cursor.execute("SELECT name FROM sqlite_master WHERE type='table';").fetchall()]
+            elif self.db_type == 'ACCESS':
+                tables = [tup[2] for tup in existing_cursor.tables() if tup[3] == 'TABLE']
+            else:
+                tables = existing_cursor.tables()
         else:
-            tables = cursor.tables()
+            with self as cursor:
+                if self.db_type == 'SQLITE':
+                    tables = [tup[0] for tup in cursor.execute("SELECT name FROM sqlite_master WHERE type='table';").fetchall()]
+                elif self.db_type == 'ACCESS':
+                    tables = [tup[2] for tup in cursor.tables() if tup[3] == 'TABLE']
+                else:
+                    tables = cursor.tables()
         return sorted(tables)
 
 
     @lru_cache(maxsize=1)
-    def query_names(self) -> list:
+    def query_names(self, existing_cursor=None) -> list:
         '''
         Return sorted list of all queries in the database.
         Only works for Access Select queries.
         '''
         if self.db_type == 'ACCESS':
-            conn, cursor = self.connection(also_cursor=True)
-            return sorted(tup[2] for tup in cursor.tables() if tup[3] == 'VIEW')
+            if existing_cursor:
+                return sorted(tup[2] for tup in existing_cursor.tables() if tup[3] == 'VIEW')
+            else:
+                with self as cursor:
+                    return sorted(tup[2] for tup in cursor.tables() if tup[3] == 'VIEW')
         else:
             return []
 
@@ -295,39 +396,35 @@ class DataBase():
         '''
         Return dict of all column: type pairs in specified table.
         '''
-        if self.db_type == 'ACCESS':
-            conn, cursor = self.connection(also_cursor=True)
-            try:
-                return {col[3]: col[5].lower() for col in cursor.columns(table=tablename)}
-            except UnicodeDecodeError:
-                print('\nERROR - Unable to read columns.')
-                print('This may occur if using Access database with column descriptions populated.')
-                print('Try deleting the column descriptions.\n')
-                return {}
-        elif self.db_type == 'SQLITE':
-            conn, cursor = self.connection(also_cursor=True)
-            return {col[1]: col[2].lower() for col in cursor.execute(f"PRAGMA TABLE_INFO('{tablename}');").fetchall()}
-        else:
-            sql = f'SELECT * FROM {tablename} LIMIT 2;'
-            conn, cursor = self.connection(also_cursor=True)
-            data = util.list_of_dicts_from_query(cursor, sql, tablename, self.db_type)
-            conn.close()
-            if len(data) == 0:
-                print(f'No rows in {tablename}.  Please determine columns and types with another method.')
-                return {}
+        with self as cursor:
+            if self.db_type == 'ACCESS':
+                try:
+                    return {col[3]: col[5].lower() for col in cursor.columns(table=tablename)}
+                except UnicodeDecodeError:
+                    print(f'\nERROR - Unable to read columns for table: {tablename}.')
+                    print('This may occur if using Access database with column descriptions populated.')
+                    print('Try deleting the column descriptions.\n')
+                    return {}
+            elif self.db_type == 'SQLITE':
+                return {col[1]: col[2].lower() for col in cursor.execute(f"PRAGMA TABLE_INFO('{tablename}');").fetchall()}
             else:
-                return {key: type(value).__name__.lower() for key, value in data[0].items()}
+                sql = f'SELECT * FROM {tablename} LIMIT 2;'
+                data = util.list_of_dicts_from_query(cursor, sql, tablename, self.db_type)
+                if len(data) == 0:
+                    print(f'No rows in {tablename}.  Please determine columns and types with another method.')
+                    return {}
+                else:
+                    return {key: type(value).__name__.lower() for key, value in data[0].items()}
 
 
     def key_columns(self, tablename: str) -> list:
         '''
-        Return the columns of the specified table that are primary keys.'
+        Return the columns of the specified table that are primary keys.
         '''
         if not self.db_type == 'ACCESS':
             print('ERROR!  .key_columns is only currently implemented for Access databases.')
         with self as cursor:
-            key_cols = [row[8] for row in cursor.statistics(tablename) if row[5] and 'key' in row[5].lower()]
-        return key_cols
+            return [row[8] for row in cursor.statistics(tablename) if row[5] and 'key' in row[5].lower()]
 
 
     def create_table(self, tablename: str, columns_and_types: dict, force_overwrite: bool=False) -> None:
@@ -335,7 +432,7 @@ class DataBase():
         Create a table in the database with name "tablename"
 
         Pass in a dictionary containing column names as keys and column types as values.
-        values can be tye actual type (like int, float, etc.) or strings of those same (like 'int', 'float', etc.)
+        values can be the actual type (like int, float, etc.) or strings of those same (like 'int', 'float', etc.)
 
         force_overwrite kwarg allows to overwrite existing table if present
         (by default will NOT overwrite/change existing table.)
@@ -347,7 +444,7 @@ class DataBase():
 
         columns_and_types = {util.clean_column_name(col): v.lower() if isinstance(v, str) else v for col, v in columns_and_types.items()}  # make sure column names are good
         try:
-            column_types = ', '.join([f'{col} {type_map[v]}' for col, v in columns_and_types.items()])
+            column_types = ', '.join([f'[{col}] {type_map[v]}' for col, v in columns_and_types.items()])
         except KeyError:
             type_values = set(columns_and_types.values())
             keys_not_in_type_map = [str(type_val) for type_val in type_values if type_val not in type_map]
@@ -406,7 +503,8 @@ class DataBase():
             return dups
 
 
-    def append(self, tablename: str, data: Union[List[dict], dict], create_table_if_needed: bool=True, safe=False, clean_column_names=False, robust: bool=True) -> None:
+    def append(self, tablename: str, data: Union[List[dict], dict], create_table_if_needed: bool=True, safe=False,
+               clean_column_names=False, robust: bool=True, progressbar: bool=None) -> None:
         '''
         Append rows of data to database table.
         Create the table in the database if it doesn't exist if create_table_if_needed is True
@@ -425,11 +523,16 @@ class DataBase():
             print('No data provided to append.')
             return
 
+        if progressbar is None:
+            if len(data) < 1000:
+                progressbar = False
+            else:
+                progressbar = True
+
         if isinstance(data, dict):  # handle case of single row append by converting it to a list
             data = [data]
 
         if clean_column_names:
-            print('Cleaning column names in data to be appended.')
             for key in list(data[0].keys()):
                 for row in data:
                     row[util.clean_column_name(key)] = row.pop(key)
@@ -437,7 +540,8 @@ class DataBase():
         if tablename not in self.table_names() and create_table_if_needed:
             self.create_table(tablename, {key: type(value).__name__ for key, value in data[0].items()})
         elif tablename not in self.table_names() and not create_table_if_needed:
-            print(f'ERROR!  Table "{tablename}" does not exist in database!\nUse create_table_if_needed=True if you would like to create it.')
+            print(f'ERROR!  Table "{tablename}" does not exist in database!')
+            print('Use create_table_if_needed=True if you would like to create it.')
             return None
 
         if robust:
@@ -448,16 +552,15 @@ class DataBase():
         if data_cols != columns:
             try:
                 data = [{col: d[col] for col in columns} for d in data]
-                print('Append data column order adjusted to match db table column order.')
             except KeyError:
                 print(f'Error!  Table {tablename} columns do not match the keys of the data to be appended.')
                 print('Try setting robust=True and/or /n  set clean_column_names=True to replace " " and "/" with underscores in data keys.')
                 return
 
         if self.db_type == 'SQLITE':
-            insert_sql = f"INSERT INTO '{tablename}' ({','.join(columns)}) VALUES "
+            insert_sql = f"INSERT INTO '{tablename}' ({','.join([f'[{col}]' for col in columns])}) VALUES "
         else:
-            insert_sql = f"INSERT INTO [{tablename}] ({', '.join(columns)}) VALUES "
+            insert_sql = f"INSERT INTO [{tablename}] ({', '.join([f'[{col}]' for col in columns])}) VALUES "
         insert_many_sql = insert_sql + f"({', '.join(['?' for _ in range(len(columns))])});"
 
         # Check for potential duplicate (key) entries if Access to avoid pyodbc error and crash of whole append.
@@ -467,47 +570,81 @@ class DataBase():
             if dup_rows:
                 non_dup_data = [d for d in data if tuple(d[key] for key in key_cols) not in dup_rows]
                 skip_count = len(data) - len(non_dup_data)
-                print(f"\n{skip_count} row{'s' if skip_count > 1 else ''} {'were' if skip_count > 1 else 'was'} skipped in .append due to being primary key duplicates\n  of rows that already exist in table: {tablename}")
+                print(f"\n{skip_count} row{'s were' if skip_count > 1 else ' was'} skipped in .append due to being primary key duplicates")
+                print(f"  of rows that already exist in table: {tablename}")
                 if not non_dup_data:
                     print('No remaining data to append.')
                     return
                 print(f'The remaining {len(non_dup_data)} rows are still being appended.\n')
                 data = non_dup_data
 
+        is_sqlite = True if self.db_type == 'SQLITE' else False
         is_access = True if self.db_type == 'ACCESS' else False
+
+        math_isnan = math.isnan  # just assigning here so don't need dot lookup in below heavily-used function
         def convert_to_sql(value):
             if value is None:
                 return 'NULL'
             elif isinstance(value, str):
                 return f"'{value}'"
-            elif isinstance(value, datetime) and is_access:
-                return f'#{value}#'  # adding "#" on either end makes the Access date insert works
+            elif isinstance(value, datetime):
+                if is_sqlite:
+                    return f"'{value}'"
+                elif is_access:
+                    return f'#{value}#'  # adding "#" on either end makes the Access date insert works
+                else:
+                    return value
+            elif math_isnan(value):
+                return 'NULL'
             else:
                 return value
 
-        conn, cursor = self.connection(also_cursor=True)
-        pbar = tqdm.tqdm(total=len(data))
-        original_data_len = len(data)
-        while len(data) > 0:
-            try:
-                if safe:
-                    for row in data[-100:]:
-                        cursor.execute(insert_sql + '(' + ','.join([f'{convert_to_sql(row[col])}' for col in columns]) + ');')
-                else:
-                    cursor.executemany(insert_many_sql, [tuple(row_dict[col] for col in columns) for row_dict in data[-100:]])
-                pbar.update(100 if len(data) >= 100 else len(data))
-                data = data[:-100]
-            except sqlite3.OperationalError:  # database is locked
-                print('database locked; retrying')
-                time.sleep(random.random() / 10)
-        pbar.close()
-        conn.commit()
-        conn.close()
-        self._pull_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something has been updated
+        with self as cursor:
+            if progressbar:
+                pbar = tqdm.tqdm(total=len(data))
+            original_data_len = len(data)
+            retry_attempts = 0
+            while len(data) > 0:
+                try:
+                    if safe:
+                        for row in data[-100:]:
+                            cursor.execute(insert_sql + '(' + ','.join([f'{convert_to_sql(row[col])}' for col in columns]) + ');')
+                    else:
+                        try:
+                            cursor.executemany(insert_many_sql, [tuple(row_dict[col] for col in columns) for row_dict in data[-100:]])
+                        except (pyodbc.IntegrityError, sqlite3.InterfaceError):
+                            # this section is just intended to help debug issues with input data by printing problematic data
+                            # pyodbc.IntegrityError may occur if null value provided for index/primary key column
+                            # sqlite3.InterfaceError may occur if an unsupported data type is provided
+                            for row_dict in data[-100:]:
+                                try:
+                                    cursor.executemany(insert_many_sql, [tuple(row_dict[col] for col in columns)])
+                                except (pyodbc.IntegrityError, sqlite3.InterfaceError):
+                                    print('\n\n\n' + '-'*50 + 'ERROR!  Triggering input row shown below:')
+                                    for key, val in row_dict.items():
+                                        print(f'    {col.ljust(15)}   |   {val}')
+                                    print('-'*50 + '\n')
+                                    cursor.executemany(insert_many_sql, [tuple(row_dict[col] for col in columns)])  # call again to trigger exception messaging and exit
+
+                    if progressbar:
+                        pbar.update(100 if len(data) >= 100 else len(data))
+                    data = data[:-100]
+                except sqlite3.OperationalError as error:  # database is locked
+                    if retry_attempts < 5:
+                        retry_attempts += 1
+                        print('Database locked?  Retrying...')
+                        time.sleep(random.random() / 10)
+                    else:
+                        print(error)
+                        break
+            if progressbar:
+                pbar.close()
+
+        self._clear_pull_cache(tablename)  # clear cache for this table as want new table pull if something has been updated
         print(f'Data inserted in "{tablename}" -> {"{:,.0f}".format(original_data_len)} rows')
 
 
-    def update(self, tablename: str, match_col: str, match_val, update_col: str, update_val, progress_handler=None) -> None:
+    def update(self, tablename: str, match_col: str, match_val, update_col: str, update_val, progress_handler=None, cache_conn: bool=False) -> None:
         '''
         Update a database table with a value or values.
 
@@ -529,6 +666,9 @@ class DataBase():
         to the sqlite3 conn.set_progress_handler function that specifies
         the interval at which the callback is called. (# of SQLite instructions)
         Basically, a larger "n" value reduces the number of callbacks.
+
+        cache_conn kwarg can be used (in Access only, not used otherwise) to cache the database connection between separate update calls
+        This greatly improves performance for many updates without having to otherwise manage the connection.
         '''
         # Abort update if match or update column does not exist in the table (may be misspelled or just missing)
         table_columns = set(self.columns_and_types(tablename).keys())
@@ -537,7 +677,7 @@ class DataBase():
                 print(f'UPDATE FAILED!  Column "{col}" not in {tablename}.')
                 return
 
-        conn, cursor = self.connection(also_cursor=True)
+        conn, cursor = self.connection(also_cursor=True, cache_conn=cache_conn)
 
         if progress_handler is not None:
             if self.db_type == 'SQLITE':  # progress_handler only currently working for sqlite
@@ -545,21 +685,29 @@ class DataBase():
             else:
                 print('progress_handler is only available for use with a SQLite database.')
 
-        sql = f'UPDATE {tablename} SET {update_col}=? WHERE {match_col}=?;'  # can't pass column names in execute statement, just values
-        if isinstance(match_val, (list, tuple)):
-            if isinstance(update_val, (list, tuple)):  # many rows to update
+        sql = f'UPDATE {tablename} SET [{update_col}]=? WHERE [{match_col}]=?;'  # can't pass column names in execute statement, just values
+        if isinstance(match_val, (list, tuple)):  # Many rows to update
+            if isinstance(update_val, (list, tuple)):  #  Many values to update
                 if len(match_val) != len(update_val) and not isinstance(update_val, str):  # many rows to update with same number of values
                     print('ERROR!  The number of match values must equal the number of update values!')
                     return
                 for m_val, u_val in tqdm.tqdm(zip(match_val, update_val), total=len(match_val)):
                     cursor.execute(sql, (u_val, m_val))
-            else:  # case of many rows to update with same value
-                for m_val in tqdm.tqdm(match_val, total=len(match_val)):
-                    cursor.execute(sql, (update_val, m_val))
+            else:  # Many rows to update with the same value; run in chucks of 100 rows each
+                with tqdm.tqdm(total=len(match_val)) as pbar:
+                    while match_val:
+                        match_to_update = match_val[:100]
+                        num_updates = len(match_to_update)
+                        match_val = match_val[100:]
+                        sql = f'UPDATE {tablename} SET [{update_col}]=? WHERE [{match_col}] IN ({",".join(["?" for _ in range(num_updates)])});'
+                        cursor.execute(sql, (update_val, *match_to_update))
+                        pbar.update(num_updates)
         else:
             cursor.execute(sql, (update_val, match_val))
         conn.commit()
-        self._pull_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something's been updated
+        if not cache_conn:
+            conn.close()
+        self._clear_pull_cache(tablename)  # clear cache for this table as want new table pull if something has been updated
 
 
     def add_column(self, tablename: str, new_col: str, new_type='str') -> None:
@@ -572,7 +720,7 @@ class DataBase():
         with self as cursor:
             cursor.execute(f'ALTER TABLE {tablename} ADD COLUMN {new_col} {new_type};')
         print(f'Column {new_col} added to {tablename}.')
-        self._pull_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something's been updated
+        self._clear_pull_cache(tablename)  # clear cache for this table as want new table pull if something's been updated
         self.columns_and_types.cache_clear()
 
 
@@ -586,7 +734,7 @@ class DataBase():
             cursor.execute(f'ALTER TABLE {tablename} DROP COLUMN "{column}";')
 
         print(f'Column {column} removed from {tablename}')
-        self._pull_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something's been updated
+        self._clear_pull_cache(tablename)  # clear cache for this table as want new table pull if something's been updated
         self.columns_and_types.cache_clear()
 
 
@@ -608,7 +756,6 @@ class DataBase():
         if self.db_type == 'SQLITE':
             with self as cursor:
                 cursor.execute(f'DELETE FROM {tablename} WHERE rowid NOT IN (SELECT max(rowid) FROM {tablename} GROUP BY {", ".join(grouping_columns)})')
-            self._pull_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something's been updated
 
         elif self.db_type == 'ACCESS':
             # TODO:  Think some sort of SQL can accomplish dup deletion better than in Python... haven't figured it out yet
@@ -629,15 +776,20 @@ class DataBase():
             data = self.pull(tablename)
             existing_combos = set()
             new_data = []
+            duplicate_found = False
             for row in reversed(data):
                 row_combo = tuple(row[col] for col in grouping_columns)
                 if row_combo not in existing_combos:
                     new_data.append(row)
                     existing_combos.add(row_combo)
-            with self as cursor:
-                cursor.execute(f'DELETE * FROM {tablename};')
-            self.append(tablename, list(reversed(new_data)), safe=True, robust=False)  # UN-reverse table entries
-            self._pull_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something has been updated
+                elif row_combo in existing_combos:
+                    duplicate_found = True
+
+            if duplicate_found:  # don't delete and recreate table if no duplicates were found (no changes)
+                self.clear_table(tablename)
+                self.append(tablename, list(reversed(new_data)), safe=True, robust=False)  # UN-reverse table entries
+
+        self._clear_pull_cache(tablename)  # clear cache for this table as want new table pull if something has been updated
 
 
     def create_index(self, tablename: str, column: str, index_name: str='', unique: bool=False) -> None:
@@ -650,6 +802,20 @@ class DataBase():
             print('.create_index is currently only implemented for SQLite databases.')
 
 
+    def clear_table(self, tablename: str) -> None:
+        '''
+        Delete/clear contents of a table, but leave the table itself and the column/type structure.
+        Shortcut/convenience method.
+        '''
+        if tablename in self.table_names():
+            with self as cursor:
+                if self.db_type == 'SQLITE':
+                    cursor.execute(f'DELETE FROM {tablename};')
+                else:
+                    cursor.execute(f'DELETE * FROM {tablename};')
+            self._clear_pull_cache(tablename)  # clear cache for this table as table has been cleared
+
+
     def drop_table(self, tablename: str) -> None:
         '''
         Drop/delete the specified table from the database.
@@ -681,7 +847,7 @@ class DataBase():
             print('ERROR!  Table deletion only implemented in SQLite and Access currently.')
             return
 
-        self._pull_cache.pop(tablename, None)  # clear cache for this table as table has been dropped
+        self._clear_pull_cache(tablename)  # clear cache for this table as table has been dropped
         self.table_names.cache_clear()  # need to repull table names as just (likely) deleted one
         self.columns_and_types.cache_clear()
 
